/***** BEGIN LICENSE BLOCK *****

    FlashGot - a Firefox extension for external download managers integration
    Copyright (C) 2004-2013 Giorgio Maone - g.maone@informaction.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                             
***** END LICENSE BLOCK *****/
LAZY_INCLUDE("YoutubeSwf");

var Youtube = {

  process: function (w, id, callback, isPrivate) {
    // First try to extract the stream data from the document:
    // either from ytplayer.config.args.url_encoded_fmt_stream_map
    // or from the player's flashvars.
    // If that fails, retry with /get_video_info.
    // XXX: The problem with using /get_video_info is that it allows
    // you to bypass the "This video is age-restricted" warning. But
    // I guess it's their problem that their script doesn't check what
    // it should check. Or did they leave that backdoor on purpose?
    // Or it's not a backdoor and the response contains some kind of
    // an "age-restricted" flag that the client must check?
    // You also can bypass it with /v/VIDEO_ID and /embed/VIDEO_ID.
    w["flashgot::Youtube::process::call_cnt"] = (w["flashgot::Youtube::process::call_cnt"] || 0) + 1;
    if (w["flashgot::Youtube::process::tid"]) {
      w.clearTimeout(w["flashgot::Youtube::process::tid"]);
      delete w["flashgot::Youtube::process::tid"];
    }

    function wait_for_player() {
      delete w["flashgot::Youtube::process::tid"];

      let now = new Date().valueOf();
      if (!arguments.callee["flashgot::start_time"]) {
        arguments.callee["flashgot::start_time"] = now;
      }
      if (now - arguments.callee["flashgot::start_time"] > 5000) {
        Youtube.do_xhr(w, id, callback, isPrivate);
        return;
      }
      let data;
      let o;
      // ytplayer.config.args
      if (!data && (o = w.wrappedJSObject) && (o = o.ytplayer) && (o = o.config) && (o = o.args) && o.url_encoded_fmt_stream_map) {
        data = o;
      }
      // <embed/> flashvars
      if (!data && (o = w.document)) {
        let ids = gFlashGotService.getPref("media.YouTube.flashvars_element_id", "movie_player").split(/\s+/).map(unescape);
        let attrs = gFlashGotService.getPref("media.YouTube.flashvars_element_attr", "flashvars").split(/\s+/).map(unescape);
        for (let i = 0, len = ids.length; !data && i !== len; ++i) {
          let id = ids[i];
          if (id.length === 0) {
            continue;
          }
          let elm = o.getElementById(id);
          if (elm === null) {
            continue;
          }
          for (let j = 0, jlen = attrs.length; j !== jlen; ++j) {
            let attr = attrs[j];
            if (attr.length !== 0 && (attr = elm.getAttribute(attr))) {
              data = Youtube.parse(attr);
              break;
            }
          }
        }
      }
      if (!data) {
        // The document is loaded, but still no data - no point in waiting further
        // (e.g. /embed/VIDEO_ID documents don't have the data we're looking for).
        if (w.document && w.document.readyState && (w.document.readyState === "interactive" || w.document.readyState === "complete")) {
          Youtube.do_xhr(w, id, callback, isPrivate);
        }
        else {
          w["flashgot::Youtube::process::tid"] = w.setTimeout(arguments.callee, 200);
        }
        return;
      }
      if (Youtube.process_video_info(data, callback)) {
        delete w["flashgot::Youtube::process::call_cnt"];
        return;
      }
      Youtube.do_xhr(w, id, callback, isPrivate);
    } // wait_for_player()
    w["flashgot::Youtube::process::tid"] = w.setTimeout(wait_for_player, 0);
  },

  do_xhr: function (w, id, callback, isPrivate) {
    var call_cnt = w["flashgot::Youtube::process::call_cnt"];

    var xhr = CC["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(CI.nsIXMLHttpRequest);
    xhr.open("GET", "http://www.youtube.com/get_video_info?hl=en_US&el=detailpage&video_id="
      + encodeURIComponent(id)
      + "&sts=" + fg.getPref("media.YouTube.decode_signature_func.timestamp", -1));
    if (isPrivate) {
      fg.privatize(xhr.channel);
    }
    xhr.addEventListener("readystatechange", function (ev) {
      if (w["flashgot::Youtube::process::call_cnt"] !== call_cnt) {
        this.abort();
        return;
      };
      if (xhr.readyState !== 4) return;
      delete w["flashgot::Youtube::process::call_cnt"];
      Youtube.process_video_info(Youtube.parse(xhr.responseText || ""), callback);
    }, false);
    xhr.send(null);
  },

  parse: function (s) {
    var pairs = s.split("&");
    var res = {};
    var nv;
    for each(var p in pairs) {
      try {
        nv = p.split("=").map(function(v) { return decodeURIComponent(v.replace(/\+/g, ' ')) });
        if (!(nv[0] in res)) res[nv[0]] = nv[1];
      } catch (e) {}
    }
    return res;
  },

  process_video_info: function (data, callback) {
    if (!data || !data.url_encoded_fmt_stream_map) { return false; }
    var streams = data.url_encoded_fmt_stream_map.split(",").map(Youtube.parse);
    if (!streams.length) { return false; }
    // Preprocess the streams:
    // 1) Mark 3D videos. MediaSniffer uses .quality ("small" (240p), "medium" (360p),
    // "hd720", etc.) and .type ("video/flv", "video/mp4", etc.) for formatting menu
    // item labels and tooltips.
    // 2) Decode the stream signatures.
    // 3) Build the sort keys. Don't sort right here,
    // because the responses can arrive in any order.
    var data_title = String(data.title);
    var signature_decoder;
    try {
      signature_decoder = this.create_signature_decoder();
    } catch (x) {
      fg.log("Error compiling YouTube.decode_signature_func: " + (x.message || x) + "\n" + x.stack);
      return false;
    }
    streams = streams.filter(function(s) {
      // We're only interested in HTTP streams, no RTMP or something.
      if (!/^https?:\/\//.test(s.url)) { return false; }

      // Save the video ID, for debugging purposes.
      s["flashgot::video_id"] = data.video_id;

      // Tidy the type: "video/3gpp;+codecs="mp4v.20.3,+mp4a.40.2" -> "video/3gpp".
      s.type = s.type.split(";")[0];

      // Mark 3D videos.
      // http://en.wikipedia.org/wiki/YouTube#Quality_and_codecs
      if ((s.itag >= 82 && s.itag <= 85) || (s.itag >= 100 && s.itag <= 102)) {
        s.quality += " (3D)";
      }

      // Decode the signature.
      try {
        s["flashgot::signature"] = signature_decoder.decode( {stream: s, video_info: data, swap: Youtube.decode_signature_swap} );
      } catch (x) {
        fg.log("Error calling YouTube.decode_signature_func: " + (x.message || x) + "\n" + x.stack);
        return false;
      }

      // Build the sort key: order by quality, type.
      s["flashgot::sort_key"] = (Youtube.get_sort_idx(Youtube.QUALITY_SORT_ORDER, s.quality) << 8)
        | Youtube.get_sort_idx(Youtube.TYPE_SORT_ORDER, s.type);

      // Tidy the URL.
      if (fg.getPref("media.YouTube.stream_url.tidy", false)) {
        var extra_keep_params = fg.getPref("media.YouTube.stream_url.extra_keep_params", "")
          .split(/\s+/)
          // URL-decode.
          .map(unescape)
          // Ignore empty items.
          .filter(function(o){return o.length !== 0;});
        Youtube.tidy_stream_url(s, extra_keep_params);
      }

      return true;
    });
    signature_decoder.dispose();
    if (!streams.length) { return false; }
    callback( {title: data_title, streams: streams} );
    return true;
  }, // process_video_info()

  QUALITY_SORT_ORDER: ["hires" /*"original"*/ , "hd1080", "hd1080 (3D)", "hd720", "hd720 (3D)", "large", "large (3D)", "medium", "medium (3D)", "small", "small (3D)"],
  TYPE_SORT_ORDER: ["video/mp4", "video/x-flv", "video/webm", "video/3gpp"],
  get_sort_idx: function (arr, value) {
    // Unknown values go to the very end of the list.
    return arr.indexOf(value) & 0xff;
  },

  // Removes junk parameters (like "fexp") from the stream URL.
  // We keep the following parameters:
  // 1) those listed in the "sparams" parameter
  // 2) "sparams"
  // 3) "signature"
  // 4) "key"
  // 5) those passed in the |extra_keep_params| argument.
  tidy_stream_url: function (stream, extra_keep_params /*= null*/ ) {
    if (!extra_keep_params) { extra_keep_params = []; }
    var url = stream.url;
    if (!url) { return; }
    var qs_idx = url.indexOf("?");
    if (qs_idx === -1 || qs_idx + 1 === url.length) { return; }
    var qs = Youtube.parse(url.substr(qs_idx + 1));
    var sparams = qs["sparams"];
    if (!sparams) { return; }
    sparams = sparams.split(",");
    sparams.push("sparams", "signature", "key");
    sparams = sparams.concat(extra_keep_params);
    var remove_params = [];
    for (var p in qs) {
      if (sparams.indexOf(p) === -1) {
        remove_params.push(p);
      }
    }
    remove_params.forEach(function(s){delete qs[s];});
    var new_qs = [];
    for (var p in qs) {
      new_qs.push(encodeURIComponent(p) + "=" + encodeURIComponent(qs[p] || ""));
    }
    stream.url = url.substr(0, qs_idx + 1) + new_qs.join("&");
  }, // tidy_stream_url()

  // For YouTube, we can't rely on simple URL comparison because
  // the query strings of our parsed streams and the ones built by
  // the player can (and most likely, will) differ.
  stream_url_equals: function (url1, url2) {
    if (!this.is_stream_url(url1) || !this.is_stream_url(url2)) { return false; }

    // The |id| parameter is something like the internal video ID:
    // it's different for two different videos (those with different
    // "public" video IDs - /watch?v=VIDEO_ID), and it's the same for
    // all the streams of the same video.
    var id1 = /[?&]id=([^&]+)/.test(url1) ? RegExp.$1 : "flashgot-id-1";
    var id2 = /[?&]id=([^&]+)/.test(url2) ? RegExp.$1 : "flashgot-id-2";
    if (id1 !== id2) { return false; }

    // The only thing that allows us to identify a stream is the |itag|
    // parameter, which is a combination of the container type (MP4,
    // FLV, etc.) and the quality (1080p, 1080p 3D, 720p, etc.).
    var itag1 = /[?&]itag=([^&]+)/.test(url1) ? RegExp.$1 : "flashgot-itag-1";
    var itag2 = /[?&]itag=([^&]+)/.test(url2) ? RegExp.$1 : "flashgot-itag-2";
    if (itag1 !== itag2) { return false; }

    // Different seek positions are considered different streams (in case
    // the user wants to save the stream from a particular position).
    var begin1 = /[?&]begin=([^&]+)/.test(url1) ? RegExp.$1 : "0";
    var begin2 = /[?&]begin=([^&]+)/.test(url2) ? RegExp.$1 : "0";
    return begin1 === begin2;
  },

  is_stream_url: function (url) {
    return /^[a-z]+:\/\/([^\/]+)\/videoplayback\?/.test(url) && /\.youtube\.com$/.test(RegExp.$1);
  },

  decode_signature: function (params) {
    /* Not encoded. */
    return params.stream.sig || "";
  },
  decode_signature_swap: function (a, idx) {
    var tmp = a[0];
    a[0] = a[idx % a.length];
    a[idx] = tmp;
    return a;
  },

  create_signature_decoder: function () {
    var s = fg.getPref("media.YouTube.decode_signature_func", "");
    if (!s) {
      return new SignatureDecoder(Youtube.decode_signature);
    }
    // Fail fast: try to compile right now to check the code for
    // syntax errors, so that we don't do all that heavy stuff for
    // sandbox initialization only to fail later in evalInSandbox()
    // and have an incorrect error message saying "error _calling_
    // the function" while actually we failed to compile it.
    var func = new Function("params", s);
    if (fg.getPref("media.YouTube.decode_signature_func.sandbox", true)) {
      // Wrap the code into a function invocation because we promised
      // to call it as a function with one parameter.
      s = "(function(params){\n" + s + "\n})(params);";
      return new SandboxedSignatureDecoder(s);
    }
    return new SignatureDecoder(func);
  },


  refresh_signature_func: function (w, callback) {
    w = w.wrappedJSObject;
    var swf_url;
    var o;
    // ytplayer.config.url
    if (w && (o = w.ytplayer) && (o = o.config)) {
      swf_url = o.url;
    }
    // yt.config_["PLAYER_CONFIG"].url
    else if (w && (o = w.yt) && (o = o.config_) && (o = o.PLAYER_CONFIG)) {
      swf_url = o.url;
    }
    if (!swf_url) { return; }
    var stream_ctx = {
      swf_url: swf_url,
      file: swf_url.split("/").pop().replace(/\?.*$/, "").replace(/#.*$/, ""),
      bytes: "",
      contentLength: -1,
      bstream: null
    };
    var stream_listener = {
      onDataAvailable: function (req, ctx, stream, offset, count) {
        stream_ctx.bstream.setInputStream(stream);
        stream_ctx.bytes += stream_ctx.bstream.readBytes(count);
      },
      onStartRequest: function (req /*, ctx*/) {
        var channel = req.QueryInterface(Components.interfaces.nsIChannel);
        if (!((channel instanceof Components.interfaces.nsIHttpChannel)
          && Components.isSuccessCode(channel.status)
          && channel.responseStatus === 200))
        {
          throw new Error("cancel"); //req.cancel(NS_BINDING_ABORTED);
        }
        stream_ctx.contentLength = channel.contentLength || -1;
        stream_ctx.bstream = Components.classes["@mozilla.org/binaryinputstream;1"]
          .createInstance(Components.interfaces.nsIBinaryInputStream);
      },
      onStopRequest: function (req, ctx, status) {
        stream_ctx.bstream = null;
        if (Components.isSuccessCode(status)) {
          if (stream_ctx.contentLength === -1 || stream_ctx.bytes.length === stream_ctx.contentLength) {
            Youtube.refresh_signature_func0(stream_ctx, callback);
          }
          else {
            fg.log("SWF content length mismatch: expected " + stream_ctx.contentLength + ", got " + stream_ctx.contentLength);
          }
        }
        else {
          fg.log("Failed to download the SWF: status=" + status);
        }
        stream_ctx = null;
      }
    }; // stream_listener
    Components.classes["@mozilla.org/network/io-service;1"]
      .getService(Components.interfaces.nsIIOService)
      .newChannel(swf_url, null, null)
      .asyncOpen(stream_listener, null);
  },

  refresh_signature_func0: function (ctx, callback) {
    ctx.log = function(msg){fg.log(msg);};
    try {
      var rc = YoutubeSwf.refresh_signature_func(ctx);
      if (!(rc && rc.func_text &&
            (rc.func_text !== fg.getPref("media.YouTube.decode_signature_func") ||
             fg.getPref("media.YouTube.decode_signature_func.timestamp") != rc.timestamp))) {
        return;
      }
      fg.log("New signature function:\n" + rc.func_text);
      fg.log("New timestamp:\n" + rc.timestamp);
      fg.setPref("media.YouTube.decode_signature_func", rc.func_text);
      fg.setPref("media.YouTube.decode_signature_func.timestamp", rc.timestamp);
      if (typeof callback === "function") callback(rc);
    } catch (x) {
      fg.log("Error refreshing signature function: " + (x.message + x) + "\n" + x.stack);
    }
    ctx.log = null;
    ctx.bytes = null;
  }
};


// class SignatureDecoder {
//   SignatureDecoder(Function func);
//   virtual string decode(Params params);
//   virtual void dispose();
// }
// class Params {
//   Map<string, string> stream;
//   Map<string, string> video_info;
//   Function swap;
// }
function SignatureDecoder(func) {
  this.func = func;
}

SignatureDecoder.prototype = {
  decode: function(params) { return this.func(params); },
  dispose: function() { this.func = null; }
};



// class SandboxedSignatureDecoder extends SignatureDecoder;
// public SandboxedSignatureDecoder(String code_str);
function SandboxedSignatureDecoder(code_str) {
  this.code_str = code_str;

  this.sandbox = new Components.utils.Sandbox(this.SANDBOX_PRINCIPAL, this.SANDBOX_OPTIONS);

  // If you simply assign objects to the sandbox properties, they will
  // be visible, but their properties won't.
  // Primitive values and functions can be assigned as-is. (Although
  // converting a function to string will return the "standard" string
  // "function() { [native code] }" because none of its properties,
  // including the toString method, will be visible.)
  //   var obj = {stream: s, video_info: data, swap: ... };
  //   sandbox.obj = obj;
  //   sandbox.str = "hi";
  //   sandbox.num = 42;
  //   sandbox.fun = function(){};
  //   // Somewhere in decode_signature_func's global scope,
  //   // i.e. |this| points to |sandbox|:
  //   this.obj; // defined.
  //   this.obj.stream; // undefined.
  //   this.str; // defined.
  //   this.num; // defined.
  //   this.fun; // defined.
  this.sb_params = this.export_to_sandbox({
    stream: null,
    video_info: null,
    swap: Youtube.decode_signature_swap
  });
  this.sandbox.params = this.sb_params;

} // SandboxedSignatureDecoder()


SandboxedSignatureDecoder.prototype = {
  // https://developer.mozilla.org/en-US/docs/Security_check_basics:
  // The null principal (whose contract ID is @mozilla.org/nullprincipal;1)
  // fails almost all security checks. It has no privileges and can't be
  // accessed by anything but itself and chrome. They aren't same-origin
  // with anything but themselves.
  SANDBOX_PRINCIPAL: Components.classes["@mozilla.org/nullprincipal;1"]
    .createInstance(Components.interfaces.nsIPrincipal),

  SANDBOX_OPTIONS: {wantComponents: false, wantXHRConstructor: false},


  // Makes all the own properties of an object available for reading
  // in a sandbox.
  // Not recursive, i.e. doesn't go in depth, i.e. if a property
  // is an object as well, then its properties won't be processed.
  // Exposing actual objects and not their copies is safe because
  // the sandboxed code can't define getters/setters for their
  // properties because we expose them as read-only, nor can it
  // assign a watcher for a property because calling watch()
  // on the exposed objects throws "can't watch non-native
  // objects of class Proxy".
  //
  // [quote src="https://developer.mozilla.org/en/docs/XPConnect_wrappers#XPCChromeObjectWrapper"]
  // XPCChromeObjectWrapper. A chrome object wrapper, or COW, is automatically
  // created when a chrome object is exposed to content. If the chrome object
  // has an __exposedProps__ property, then content may only access the whitelisted
  // properties. This allows chrome to expose limited parts of objects to content
  // without having to use IDL.
  // The __exposedProps__ property is an object with a property corresponding
  // to each of the chrome object's properties which should be exposed. The value
  // of the properties within __exposedProps__ is either "w", "wr", or "r". These
  // will allow content to read and/or write the property (executing a function
  // is considered reading). Remember that functions defined in chrome context
  // execute with chrome privileges.
  // Gecko 15.0 note (Firefox 15.0 / Thunderbird 15.0 / SeaMonkey 2.12):
  //   Starting in Gecko 15.0 (Firefox 15.0 / Thunderbird 15.0 / SeaMonkey 2.12),
  //   exposing functions or objects without using __exposedProps__ is deprecated.
  //   Using __exposedProps__ will be mandatory starting with Gecko 17.0 (Firefox 17.0
  // / Thunderbird 17.0 / SeaMonkey 2.14 / Firefox OS 1.0).
  // [/quote]
  export_to_sandbox: function (obj) {
    var props = {"toString": "r", "valueOf": "r"};
    for (var p in obj) {
      if (p !== "__exposedProps__" && obj.hasOwnProperty(p)) {
        props[p] = "r";
      }
    }
    obj.__exposedProps__ = props;
    return obj;
  }, // export_to_sandbox()


  decode: function (params) {
    this.sb_params.stream = this.export_to_sandbox(params.stream);

    // First call - export the video_info.
    if (!this.sb_params.video_info) {
      this.sb_params.video_info = this.export_to_sandbox(params.video_info);
    }

    var rc = Components.utils.evalInSandbox(this.code_str, this.sandbox);
    delete params.stream.__exposedProps__;
    // No fancy return values - we expect a primitive string value.
    // We don't silently return something that could pass for a signature.
    // Instead, we throw - to inform the user that their decode_signature_func
    // function is broken (anyone can make a typo) or malicious.
    //
    // It's OK to pass uncaught exceptions as-is because even if they have
    // getters for properties like "message", those will be executed in the
    // context of the sandbox (i.e. the global |this| will point to the sandbox),
    // which is useless for malicious code anyway.
    // Here's what am I talking about - somewhere in the sandboxed code:
    //   var x = new Error();
    //   x.__defineGetter__("message", function(){alert("pwned");});
    //   throw x;
    //   or:
    //   var x = { message: { valueOf: function(){alert("pwned");}, toString: function(){alert("pwned");} } };
    //   throw x;
    // We could catch the exceptions here, manually sanitize them and rethrow
    // if they're safe to use in our chrome code, but I just don't see the point
    // in doing so because if there's a bug in the security manager, then our
    // manual sanitization will just conceal it.
    if (typeof (rc) === "string") { return rc; }
    // Nulls are kinda OK.
    if (rc === null) { return ""; }
    // A forgotten return or outdated code that returns nonexistent stream
    // properties? Worth a warning in either case.
    if (rc === undefined) {
      fg.log("WARNING: YouTube.decode_signature_func returned undefined.");
      return "";
    }
    throw new Error("Invalid return value type: expected string, got " + typeof (rc));
  }, // decode()

  dispose: function () {
    if (!this.sandbox) { return; }
    this.sandbox.params = this.sb_params = null;
    Components.utils.nukeSandbox(this.sandbox);
    this.sandbox = null;
  }
}; // SandboxedSignatureDecoder.prototype
